@@ -50,7 +50,7 @@ PG_MODULE_MAGIC;
 #define PGSYSCONF_COLS  		3
 #define PGFADVISE_COLS			4
 #define PGFADVISE_LOADER_COLS	5
-#define PGFINCORE_COLS  		8
+#define PGFINCORE_COLS  		10
 
 #define PGF_WILLNEED	10
 #define PGF_DONTNEED	20
@@ -58,6 +58,13 @@ PG_MODULE_MAGIC;
 #define PGF_SEQUENTIAL	40
 #define PGF_RANDOM		50
 
+#define FINCORE_PRESENT 0x1
+#define FINCORE_DIRTY   0x2
+#ifndef HAVE_FINCORE
+#define FINCORE_BITS    1
+#else
+#define FINCORE_BITS    2
+#endif
 /*
  * pgfadvise_fctx structure is needed
  * to keep track of relation path, segment number, ...
@@ -118,6 +125,8 @@ typedef struct
 	size_t	rel_os_pages;
 	size_t	pages_mem;
 	size_t	group_mem;
+	size_t	pages_dirty;
+	size_t	group_dirty;
 	VarBit	*databit;
 } pgfincoreStruct;
 
@@ -135,6 +144,8 @@ static int	pgfadvise_loader_file(char *filename,
 Datum		pgfincore(PG_FUNCTION_ARGS);
 static int	pgfincore_file(char *filename, pgfincoreStruct *pgfncr);
 
+Datum		pgfincore_drawer(PG_FUNCTION_ARGS);
+
 /*
  * We need to add some handler to keep the code clean
  * and support 8.3, 8.4 and 9.0
@@ -664,8 +675,8 @@ pgfadvise_loader(PG_FUNCTION_ARGS)
 
 	if (PG_ARGISNULL(5))
 		elog(ERROR, "pgfadvise_loader: databit argument shouldn't be NULL");
-	else
-		databit		= PG_GETARG_VARBIT_P(5);
+
+        databit		= PG_GETARG_VARBIT_P(5);
 
 	/* initialize nulls array to build the tuple */
 	memset(nulls, 0, sizeof(nulls));
@@ -736,6 +747,7 @@ static int
 pgfincore_file(char *filename, pgfincoreStruct *pgfncr)
 {
 	int		flag=1;
+	int		flag_dirty=1;
 
 	int		len, bitlen;
 	bits8	*r;
@@ -752,7 +764,9 @@ pgfincore_file(char *filename, pgfincoreStruct *pgfncr)
 	int	fd;
 	struct stat st;
 
+#ifndef HAVE_FINCORE
 	void 		  *pa  = (char *) 0;
+#endif
 	unsigned char *vec = (unsigned char *) 0;
 
 	/*
@@ -765,6 +779,8 @@ pgfincore_file(char *filename, pgfincoreStruct *pgfncr)
 	 */
 	pgfncr->pages_mem		= 0;
 	pgfncr->group_mem		= 0;
+	pgfncr->pages_dirty		= 0;
+	pgfncr->group_dirty		= 0;
 	pgfncr->rel_os_pages	= 0;
 
 	/*
@@ -795,7 +811,7 @@ pgfincore_file(char *filename, pgfincoreStruct *pgfncr)
 		/* number of pages in the current file */
 		pgfncr->rel_os_pages = (st.st_size+pgfncr->pageSize-1)/pgfncr->pageSize;
 
-		/* TODO We need to split mmap size to be sure (?) to be able to mmap */
+#ifndef HAVE_FINCORE
 		pa = mmap(NULL, st.st_size, PROT_NONE, MAP_SHARED, fd, 0);
 		if (pa == MAP_FAILED)
 		{
@@ -804,33 +820,44 @@ pgfincore_file(char *filename, pgfincoreStruct *pgfncr)
 			     filename, errno, strerror(errno));
 			return 3;
 		}
+#endif
 
 		/* Prepare our vector containing all blocks information */
 		vec = calloc(1, (st.st_size+pgfncr->pageSize-1)/pgfncr->pageSize);
 		if ((void *)0 == vec)
 		{
+#ifndef HAVE_FINCORE
 			munmap(pa, st.st_size);
+#endif
 			FreeFile(fp);
 			elog(ERROR, "Can not calloc object file : %s",
 			     filename);
 			return 4;
 		}
 
+#ifndef HAVE_FINCORE
 		/* Affect vec with mincore */
 		if (mincore(pa, st.st_size, vec) != 0)
 		{
-			free(vec);
 			munmap(pa, st.st_size);
-			FreeFile(fp);
 			elog(ERROR, "mincore(%p, %lld, %p): %s\n",
 			     pa, (long long int)st.st_size, vec, strerror(errno));
+#else
+		/* Affect vec with fincore */
+		if (fincore(fd, 0, st.st_size, vec) != 0)
+		{
+			elog(ERROR, "fincore(%u, 0, %lld, %p): %s\n",
+			     fd, (long long int)st.st_size, vec, strerror(errno));
+#endif
+			free(vec);
+			FreeFile(fp);
 			return 5;
 		}
 
 		/*
 		 * prepare the bit string
 		 */
-		bitlen = (st.st_size+pgfncr->pageSize-1)/pgfncr->pageSize;
+		bitlen = FINCORE_BITS * ((st.st_size+pgfncr->pageSize-1)/pgfncr->pageSize);
 		len = VARBITTOTALLEN(bitlen);
 		/*
 		 * set to 0 so that *r is always initialised and string is zero-padded
@@ -847,10 +874,24 @@ pgfincore_file(char *filename, pgfincoreStruct *pgfncr)
 		for (pageIndex = 0; pageIndex <= pgfncr->rel_os_pages; pageIndex++)
 		{
 			// block in memory
-			if (vec[pageIndex] & 1)
+			if (vec[pageIndex] & FINCORE_PRESENT)
 			{
 				pgfncr->pages_mem++;
 				*r |= x;
+				if (FINCORE_BITS > 1)
+				{
+					if (vec[pageIndex] & FINCORE_DIRTY)
+					{
+						pgfncr->pages_dirty++;
+						*r |= (x >> 1);
+						/* we flag to detect contigous blocks in the same state */
+						if (flag_dirty)
+							pgfncr->group_dirty++;
+						flag_dirty = 0;
+					}
+					else
+						flag_dirty = 1;
+				}
 				elog (DEBUG5, "in memory blocks : %lld / %lld",
 				      (long long int) pageIndex, (long long int) pgfncr->rel_os_pages);
 
@@ -862,7 +903,8 @@ pgfincore_file(char *filename, pgfincoreStruct *pgfncr)
 			else
 				flag=1;
 
-			x >>= 1;
+
+			x >>= FINCORE_BITS;
 			if (x == 0)
 			{
 				x = HIGHBIT;
@@ -877,7 +919,9 @@ pgfincore_file(char *filename, pgfincoreStruct *pgfncr)
 	 * free and close
 	 */
 	free(vec);
+#ifndef HAVE_FINCORE
 	munmap(pa, st.st_size);
+#endif
 	FreeFile(fp);
 
 	/*
@@ -1038,6 +1082,10 @@ pgfincore(PG_FUNCTION_ARGS)
 			nulls[7]  = true;
 			values[7] = (Datum) NULL;
 		}
+		/* number of pages dirty in OS cache */
+		values[8] = Int64GetDatum(pgfncr->pages_dirty);
+		/* number of group of contigous dirty pages in os cache */
+		values[9] = Int64GetDatum(pgfncr->group_dirty);
 		/* Build the result tuple. */
 		tuple = heap_form_tuple(fctx->tupd, values, nulls);
 
@@ -1048,3 +1096,70 @@ pgfincore(PG_FUNCTION_ARGS)
 		SRF_RETURN_NEXT(funcctx, HeapTupleGetDatum(tuple));
 	}
 }
+
+/*
+ * pgfincore_drawer A very naive renderer. (for testing)
+ */
+PG_FUNCTION_INFO_V1(pgfincore_drawer);
+Datum
+pgfincore_drawer(PG_FUNCTION_ARGS)
+{
+        char *result,
+             *r;
+	int  len,i,k;
+	VarBit *databit;
+	bits8 *sp;
+	bits8 x;
+
+	if (PG_ARGISNULL(1))
+		elog(ERROR, "pgfincore_drawer: databit argument shouldn't be NULL");
+
+        databit	= PG_GETARG_VARBIT_P(0);
+
+	len =  VARBITLEN(databit);
+	result = (char *) palloc((len/FINCORE_BITS) + 1);
+	sp = VARBITS(databit);
+	r = result;
+
+	for (i = 0; i <= len - BITS_PER_BYTE; i += BITS_PER_BYTE, sp++)
+	{
+		x = *sp;
+		/*  Is this bit set ? */
+		for (k = 0; k < (BITS_PER_BYTE/FINCORE_BITS); k++)
+		{
+		  char out = ' ';
+			if (IS_HIGHBIT_SET(x))
+			  out = '.' ;
+			x <<= 1;
+			if (FINCORE_BITS > 1)
+			{
+				if (IS_HIGHBIT_SET(x))
+					out = '*';
+				x <<= 1;
+			}
+			*r++ = out;
+		}
+	}
+	if (i < len)
+	{
+		/* print the last partial byte */
+		x = *sp;
+		for (k = i; k < (len/FINCORE_BITS); k++)
+		{
+		        char out = ' ';
+			if (IS_HIGHBIT_SET(x))
+			  out = '.' ;
+			x <<= 1;
+			if (FINCORE_BITS > 1)
+			{
+				if (IS_HIGHBIT_SET(x))
+					out = '*';
+				x <<= 1;
+			}
+			*r++ = out;
+		}
+	}
+
+	*r = '\0';
+	PG_RETURN_CSTRING(result);
+}
